/* bsgs.c -- (C) Geoffrey Reynolds, April 2006.

   Implementation of a baby step giant step algorithm for finding all
   n in the range nmin <= n <= nmax satisfying b^n=d_i (mod p) where b
   and each d_i are relatively prime to p.

   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2 of the License, or
   (at your option) any later version.
*/

#include <assert.h>
#include <inttypes.h>
#include <math.h>
#include <stdlib.h>
#include "sr1sieve.h"
#include "arithmetic.h"
#include "bitmap.h"
#include "hashtable.h"

/*
  Giant step baby step algorithm, from Wikipaedea:

  input: A cyclic group G of order n, having a generator a and an element b.

  Output: A value x satisfying a^x = b (mod n).

  1. m <-- ceiling(sqrt(n))
  2. For all j where 0 <= j < m:
     1. Compute a^j mod n and store the pair (j, a^j) in a table.
  3. Compute a^(-m).
  4. g <-- b.
  5. For i = 0 to (m - 1):
     1. Check if g is the second component (a^j) of any pair in the table.
     2. If so, return im + j.
     3. If not, g <-- ga^(-m) mod n.
*/


#ifndef HAVE_invmod64
#define HAVE_invmod64
/*
  1/a (mod p). Assumes 1 < a < 2^32, a < p.
*/
static uint64_t invmod32_64(uint32_t a, uint64_t p)
{
  /* Thanks to the folks at mersenneforum.org.
     See http://www.mersenneforum.org/showthread.php?p=58252. */

  uint64_t ps1, ps2, q;
  uint32_t r, t, dividend, divisor;
  int sign = -1;

  assert(1 < a);
  assert(a < p);

  q = p / a;
  r = p % a;

  assert(r > 0);

  dividend = a;
  divisor = r;
  ps1 = q;
  ps2 = 1;
	
  while (divisor > 1)
  {
    r = dividend - divisor;
    t = r - divisor;
    if (r >= divisor) {
      q += ps1; r = t; t -= divisor;
      if (r >= divisor) {
        q += ps1; r = t; t -= divisor;
        if (r >= divisor) {
          q += ps1; r = t; t -= divisor;
          if (r >= divisor) {
            q += ps1; r = t; t -= divisor;
            if (r >= divisor) {
              q += ps1; r = t; t -= divisor;
              if (r >= divisor) {
                q += ps1; r = t; t -= divisor;
		if (r >= divisor) {
                  q += ps1; r = t; t -= divisor;
                  if (r >= divisor) {
                    q += ps1; r = t;
                    if (r >= divisor) {
                      q = dividend / divisor;
                      r = dividend % divisor;
                      q *= ps1;
                    } } } } } } } } }
    q += ps2;
    sign = -sign;
    dividend = divisor;
    divisor = r;
    ps2 = ps1;
    ps1 = q;
  }

  assert(0 < ps1);
  assert(ps1 < p);

  return (sign > 0) ? ps1 : p - ps1;
}

/*
  1/a (mod p) if sign is negative or -1/a (mod p) otherwise.
  Assumes 1 < a < p.
*/
static uint64_t invmod64_64(uint64_t a, int32_t sign, uint64_t p)
{
  /* Thanks to the folks at mersenneforum.org.
     See http://www.mersenneforum.org/showthread.php?p=58252. */

  uint64_t ps1, ps2, q, r, t, dividend, divisor;

  assert(1 < a);
  assert(a < p);
  assert(sign != 0);

  q = p / a;
  r = p % a;

  assert(r > 0);

  dividend = a;
  divisor = r;
  ps1 = q;
  ps2 = 1;
	
  while (divisor > 1)
  {
    r = dividend - divisor;
    t = r - divisor;
    if (r >= divisor) {
      q += ps1; r = t; t -= divisor;
      if (r >= divisor) {
        q += ps1; r = t; t -= divisor;
        if (r >= divisor) {
          q += ps1; r = t; t -= divisor;
          if (r >= divisor) {
            q += ps1; r = t; t -= divisor;
            if (r >= divisor) {
              q += ps1; r = t; t -= divisor;
              if (r >= divisor) {
                q += ps1; r = t; t -= divisor;
		if (r >= divisor) {
                  q += ps1; r = t; t -= divisor;
                  if (r >= divisor) {
                    q += ps1; r = t;
                    if (r >= divisor) {
                      q = dividend / divisor;
                      r = dividend % divisor;
                      q *= ps1;
                    } } } } } } } } }
    q += ps2;
    sign = -sign;
    dividend = divisor;
    divisor = r;
    ps2 = ps1;
    ps1 = q;
  }

  assert(0 < ps1);
  assert(ps1 < p);

  return (sign > 0) ? ps1 : p - ps1;
}
#endif


/* Pre-compute the largest power bdQ=b^(Q/dQ) < p_min, where dQ divides Q.
   This allows b^Q (mod p) to be computed later as bdQ^dQ (mod p).
*/
#define PRE_POWER 1

static uint32_t m;     /* Number of baby steps */
static uint32_t M;     /* Number of giant steps */
static uint32_t sieve_low;

typedef struct {uint32_t m; uint32_t M;} steps_t;
static steps_t *steps;

#if PRE_POWER
static uint32_t dQ;           /* Minimum dQ|Q such that b^(Q/dQ)<p_min */
static uint64_t bdQ;          /* b^(Q/dQ) */
#endif

#if USE_VECTOR
static uint32_t attribute((noinline)) baby_steps64(uint64_t p)
{
  v128_t B,P;
  uint64_t b, bj0;
  uint32_t j;

  /* b <- base^Q (mod p) */
#if PRE_POWER
  b = powmod64(bdQ,dQ,p);
#else
  b = powmod64(b_term,subseq_Q,p);
#endif

  bj0 = powmod64(b,sieve_low,p);
  BJ64[0] = bj0;
  PRE2_MULMOD64_INIT(b);
  for (j = 0; j < 15; j++)
    BJ64[j+1] = PRE2_MULMOD64(BJ64[j],b,p);
  PRE2_MULMOD64_FINI();

  b = powmod64(b,16,p);
  B = vec2_set2(b);
  P = vec2_set2(p);

  PRE2_MULMOD64_INIT(b);
  for (j = 0; j+16 < m; j += 16)
    VEC16_MULMOD64_NEXT16(BJ64+j,B,P);
  PRE2_MULMOD64_FINI();
  BJ64[m] = bj0;

  clear_hashtable();
  j = 0;
  do { post_insert64(j); }
  while (BJ64[++j] != bj0);

  return (j < m) ? j : 0;
}
#else
static uint32_t baby_steps64(uint64_t p)
{
  uint64_t b, bj, bj0;
  uint32_t j;

  /* b <- base^Q (mod p) */
#if PRE_POWER
  b = powmod64(bdQ,dQ,p);
#else
  b = powmod64(b_term,subseq_Q,p);
#endif

  bj0 = powmod64(b,sieve_low,p);
  clear_hashtable();
  PRE2_MULMOD64_INIT(b);
  for (j = 0, bj = bj0; j < m; j++)
  {
    insert64(j, bj);
    bj = PRE2_MULMOD64(bj,b,p);
    if (bj == bj0) /* Unlikely */
    {
      PRE2_MULMOD64_FINI();
      return j+1;
    }
  }

  PRE2_MULMOD64_FINI();
  return 0;
}
#endif

static uint64_t bQ;           /* 1/b^Q (mod p) */
static const uint16_t *C16;   /* List of candidate subsequences for BSGS */
static uint64_t *D64;         /* D64[j] holds -c/(k*b^(im+d)) (mod p) for
                                 subsequence C16[j] at giant step i. */

/* Assign BJ64[i] = 1/b^i (mod p) for each i in the ladder.
   Return 1/b^Q (mod p).
 */
static uint64_t climb_ladder(const uint16_t *ladder, uint64_t inv, uint64_t p)
{
  uint32_t i;

  BJ64[0] = 1;
  BJ64[1] = inv;
  BJ64[2] = sqrmod64(inv,p);
  for (i = 2; *ladder > 0; i += *ladder++)
    BJ64[i+*ladder] = mulmod64(BJ64[i],BJ64[*ladder],p);

  return BJ64[subseq_Q];
}

/* This function builds the list C16[] of subsequences (k*b^d)*(b^Q)^m+c for
   which p may be a factor (-ckb^d is a quadratic/cubic/quartic/quintic
   residue with respect to p) and initialises the table D64[] with the
   values -c/(k*b^d) (mod p). As a side effect, bQ is set to the value 1/b^Q
   (mod p) for use later in bsgs64(), and m,M are set to the best choice of
   baby,giant steps. Returns the number of subsequences listed in C16[].
*/
static uint32_t setup64(uint64_t p, int parity)
{
  uint64_t inv, y_k, p_s, X[POWER_RESIDUE_LCM+1];
  const uint16_t *list;
  uint32_t r, h, j;

  /* inv <-- 1/b (mod p) */
  inv = (b_term == 2) ? (p+1)/2 : invmod32_64(b_term,p);

  int16_t s = div_shift[(p/2)%(POWER_RESIDUE_LCM/2)];
  if (s == 0)
  {
    /* p = 1 (mod 2) is all we know, check for quadratic residues only.
     */

    /* Precompute 1/b^d (mod p) for 0 <= d <= Q, as necessary. */
    bQ = climb_ladder(sc_ladders[parity+1][0][0],inv,p);

    /* For each subsequence (k*b^d)*(b^Q)^(n/Q)+c, compute
       -c/(k*b^d) (mod p) and add the subsequence to the bsgs list.
    */
    inv = invmod64_64(k_term,c_term,p); /* inv <-- -c/k (mod p) */
    PRE2_MULMOD64_INIT(inv);
    C16 = sc_lists[parity+1][0][0];
    for (j = 0, list = C16; (h = *list++) < SUBSEQ_MAX; j++)
      D64[j] = PRE2_MULMOD64(BJ64[SUBSEQ[h].d],inv,p); /* -c/(k*b^d) */
    PRE2_MULMOD64_FINI();
    m = steps[j].m;
    M = steps[j].M;
    return j;
  }
  else if (s > 0)
  {
    /* p = 1 (mod s), where s is not a power of 2. Check for r-th power
       residues for each prime power divisor r of s.
    */
    p_s = p/s;
  }
  else /* s < 0 */
  {
    /* p = 1 (mod 2^s), where s > 1. Check for r-th power residues for each
       divisor r of s. We handle this case seperately to avoid computing p/s
       using plain division.
    */
    p_s = p >> (-s);
#ifndef NDEBUG
    s = 1 << (-s);
#endif
  }

  /* For 0 <= r < s, X[r] <- 1/(b^r)^((p-1)/s) */
  X[0] = 1;
  X[1] = powmod64(inv,p_s,p);
  PRE2_MULMOD64_INIT(X[1]);
  for (r = 1; X[r] != 1; r++)
    X[r+1] = PRE2_MULMOD64(X[r],X[1],p);
  PRE2_MULMOD64_FINI();
  assert(s%r == 0);
  /* 1/(b^r)^((p-1)/s)=1 (mod p) therefore (1/(b^r)^((p-1)/s))^y=1 (mod p)
     for 0 <= y < s/r. (Could we do more with this?)
  */

  /* y_k <- (-ck)^((p-1)/s) */
  y_k = powmod64((c_term == -1) ? k_term : p-k_term, p_s, p);

  for (h = 0; h < r; h++)
    if (y_k == X[h]) /* i.e. (-ckb^h)^((p-1)/r)=1 (mod p) */
      break;

  if (h < r && (C16 = sc_lists[parity+1][divisor_index[r]][h]) != NULL)
  {
    /* -ckb^n is an r-power residue for at least one term k*b^n+c of
       this sequence.
    */

    /* Precompute 1/b^d (mod p) for 0 <= d <= Q, as necessary. */
    bQ = climb_ladder(sc_ladders[parity+1][divisor_index[r]][h],inv,p);
    inv = invmod64_64(k_term,c_term,p); /* inv <-- -c/k (mod p) */
    PRE2_MULMOD64_INIT(inv);
    for (j = 0, list = C16; (h = *list++) < SUBSEQ_MAX; j++)
    {
      /* -ckb^d is an r-th power residue for at least one term
         (k*b^d)*(b^Q)^(n/Q)+c of this subsequence.
      */
      D64[j] = PRE2_MULMOD64(BJ64[SUBSEQ[h].d],inv,p); /* -c/(k*b^d) */
    }
    PRE2_MULMOD64_FINI();
    m = steps[j].m;
    M = steps[j].M;
    return j;
  }

  return 0;
}

static void bsgs64(uint64_t p, uint32_t cc)
{
  uint32_t i, j, k;

  assert(cc > 0);

  /* Baby steps. */
  if ((i = baby_steps64(p)) > 0)
  {
    /* i is the order of b (mod p). This is all the information we need to
       determine every solution for this p, so no giant steps are needed. */

    for (k = 0; k < cc; k++)
      for (j = lookup64(D64[k]); j < m*M; j += i)
        eliminate_term(C16[k],sieve_low+j,p);
  }
  else
  {
    /* First giant step. */
    for (k = 0; k < cc; k++)
      if ((j = lookup64(D64[k])) != HASH_NOT_FOUND)
        eliminate_term(C16[k],sieve_low+j,p);

    /* Remaining giant steps. */
    if (M > 1)
    {
      uint64_t b;
#if USE_VECTOR
      v128_t B,P;
      b = powmod64(bQ,m,p); /* b <- 1/b^Qm (mod p) */
      B = vec2_set2(b);
      P = vec2_set2(p);
      PRE2_MULMOD64_INIT(b);
      for (i = 1; i < M; i++)
      {
        for (k = 0; k < cc; k += 8)
          VEC8_MULMOD64_THIS8(D64+k,B,P);
        for (k = 0; k < cc; k++)
          if ((j = lookup64(D64[k])) != HASH_NOT_FOUND)
            eliminate_term(C16[k],sieve_low+i*m+j,p);
      }
      PRE2_MULMOD64_FINI();
#else
      b = powmod64(bQ,m,p); /* b <- 1/b^Qm (mod p) */
      PRE2_MULMOD64_INIT(b);
      for (i = 1; i < M; i++)
        for (k = 0; k < cc; k++)
        {
          D64[k] = PRE2_MULMOD64(D64[k],b,p);
          if ((j = lookup64(D64[k])) != HASH_NOT_FOUND)
            eliminate_term(C16[k],sieve_low+i*m+j,p);
        }
      PRE2_MULMOD64_FINI();
#endif
    }
  }
}

static void choose_steps(uint32_t r, uint32_t s)
{
  uint32_t m, M, max_elts;

  M = MAX(1,rint(sqrt((double)r/s)));
  m = MIN(r,ceil((double)r/M));

#if HASHTABLE_SIZE_OPT
  max_elts = (hashtable_size) ? hashtable_size/sizeof(HASH_T) : HASH_MAX_ELTS;
#else
  max_elts = HASH_MAX_ELTS;
#endif

  if (m > max_elts)
  {
    /* There are three ways to handle this undesirable case:
       1. Allow m > HASH_MAX_ELTS (undersize hash table).
       2. Restrict m <= HASH_MAX_ELTS (suboptimal baby/giant-step ratio).
       3. Recompile with SHORT_HASHTABLE=0.
    */
    M = ceil((double)r/max_elts);
    m = ceil((double)r/M);
  }

  assert(m <= max_elts);

  steps[s].m = m;
  steps[s].M = M;
}

static void init_bsgs(void)
{
  uint32_t r, s, sieve_high, max_baby_steps;

  sieve_low = n_min / subseq_Q;
  sieve_high = n_max / subseq_Q;

  r = sieve_high - sieve_low + 1;

  /*
    r = range of n, s = number of subsequences.
    In the worst case we will do do one table insertion and one mulmod
    for m baby steps, then s table lookups and s mulmods for M giant
    steps. The average case depends on how many solutions are found
    and how early in the loop they are found, which I don't know how
    to analyse. However for the worst case we just want to minimise
    m + s*M subject to m*M >= r, which is when m = sqrt(s*r).
  */

  steps = xmalloc((subseq_count+1)*sizeof(steps_t));

  for (s = 1, max_baby_steps = 0; s <= subseq_count; s++)
  {
    choose_steps(r,s);
    assert(sieve_high < sieve_low+steps[s].m*steps[s].M);
    max_baby_steps = MAX(max_baby_steps,steps[s].m);
  }

#if USE_VECTOR
  /* Allow room for vector operations to overrun the end of the array.
   */
  D64 = xmemalign16((subseq_count+15)*sizeof(uint64_t));
#else
  D64 = xmemalign16((subseq_count+1)*sizeof(uint64_t));
#endif
  init_hashtable(max_baby_steps);

#if PRE_POWER
  {
    uint32_t i;
    uint64_t bi;

    dQ = subseq_Q;
    bdQ = b_term;
    for (i=2, bi=b_term; i <= subseq_Q && bi < p_min/b_term; i++)
    {
      bi *= b_term;
      if (subseq_Q % i == 0)
      {
        dQ = subseq_Q/i;
        bdQ = bi;
      }
    }
  }
#endif
}

static void fini_bsgs(void)
{
  fini_hashtable();
  xfree16(D64);
  free(steps);
}

void CODE_PATH_NAME(sieve)(void)
{
  init_bsgs();
  init_prime_sieve(p_max);
  start_srsieve();
#if CHECK_FOR_GFN
  if (seq_gfn)
    prime_sieve_gfn(p_min,p_max,setup64,bsgs64,seq_gfn);
  else
    prime_sieve(p_min,p_max,setup64,bsgs64);
#else
  prime_sieve(p_min,p_max,setup64,bsgs64);
#endif
  finish_srsieve("range is complete",p_max);
  fini_prime_sieve();
  fini_bsgs();
}
