/* sr1sieve.c -- (C) Geoffrey Reynolds, June 2006.

   A version of srsieve specialised for a single sequence k*b^n+/-1.

   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2 of the License, or
   (at your option) any later version.
*/

#include <assert.h>
#include <inttypes.h>
#include <math.h>
#include <signal.h>
#include <stdlib.h>
#include <stdio.h>
#include <time.h>
#include <getopt.h>
#include <errno.h>
#include "sr1sieve.h"
#include "version.h"
#include "arithmetic.h"

#if HAVE_MALLOPT
#include <malloc.h>
#endif

#define NAME "sr1sieve"
#define DESC "A sieve for one sequence k*b^n+/-1"

const char *output_file_name = NULL;
const char *factors_file_name = NULL;
uint32_t save_period = DEFAULT_SAVE_PERIOD;
int verbose_opt = 0;
uint32_t L1_cache_size = 0;
uint32_t L2_cache_size = 0;
#if HASHTABLE_SIZE_OPT
uint32_t hashtable_size = 0;
#endif

static void banner(void)
{
  printf("%s %d.%d.%d -- %s.\n", NAME, MAJOR_VER, MINOR_VER, PATCH_VER, DESC);
}

static void help(void)
{
  printf("Usage: %s -P P1 -i FILE <-o FILE | -f FILE> [OPTION ...]\n", NAME);
  printf(" -p --pmin P0\n");
  printf(" -P --pmax P1          "
         "Sieve for factors p in the range P0 <= p <= P1\n");
  printf(" -i --input FILE       "
         "Read sieve from NewPGen format file FILE.\n");
  printf(" -o --output FILE      "
         "Write sieve to NewPGen format file FILE.\n");
  printf(" -f --factors FILE     "
         "Append new factors to file FILE.\n");
#if LEGENDRE_CACHE
  printf(" -C --cache-file FILE  "
         "Load (or save) Legendre symbol tables from (or to) FILE.\n");
#endif
  printf(" -s --save TIME        "
         "Update output file every TIME (default %u) minutes.\n",
         (unsigned int)DEFAULT_SAVE_PERIOD/60);
#if MULTI_PATH
  printf("    --sse2             "
         "Use SSE2 vector optimisations.\n");
  printf("    --no-sse2          "
         "Don't use SSE2 vector optimisations.\n");
#endif
  printf(" -l --L1-cache SIZE    "
         "Assume L1 data cache is SIZE Kb.\n");
  printf(" -L --L2-cache SIZE    "
         "Assume L2 cache is SIZE Kb.\n");
#if HASHTABLE_SIZE_OPT
  printf(" -H --hashtable SIZE   "
         "Force use of a SIZE Kb hashtable.\n");
#endif
#if SUBSEQ_Q_OPT
  printf(" -Q --subseq Q         "
         "Force sieving k*b^n+c as subsequences (k*b^d)*(b^Q)^m+c.\n");
#endif
  printf(" -x --no-lookup        "
         "Don't pre-compute Legendre symbol lookup tables.\n");
#if IDLE_OPT
#if IDLE_DEFAULT
  printf(" -z --idle             "
         "Run at idle priority. (Default)\n");
  printf(" -Z --no-idle          "
         "Don't adjust priority.\n");
#else
  printf(" -z --idle             "
         "Run at idle priority.\n");
  printf(" -Z --no-idle          "
         "Don't adjust priority. (Default)\n");
#endif
#endif
  printf(" -d --duplicates       "
         "Report factors that don't eliminate any composite.\n");
  printf(" -v --verbose          "
         "Print some extra messages.\n");
  printf(" -h --help             "
         "Print this help.\n\n");

  exit(EXIT_SUCCESS);
}

static const char *short_opts = "p:P:i:o:f:C:s:l:L:H:Q:xzZdvh";
static const struct option long_opts[] =
  {
    {"pmin",       required_argument, 0, 'p'},
    {"pmax",       required_argument, 0, 'P'},
    {"input",      required_argument, 0, 'i'},
    {"output",     required_argument, 0, 'o'},
    {"factors",    required_argument, 0, 'f'},
#if LEGENDRE_CACHE
    {"cache-file", required_argument, 0, 'C'},
#endif
    {"save",       required_argument, 0, 's'},
#if MULTI_PATH
    {"no-sse2",    no_argument,       0, '1'},
    {"sse2",       no_argument,       0, '2'},
#endif
    {"L1-cache",   required_argument, 0, 'l'},
    {"L2-cache",   required_argument, 0, 'L'},
#if HASHTABLE_SIZE_OPT
    {"hashtable",  required_argument, 0, 'H'},
#endif
#if SUBSEQ_Q_OPT
    {"subseq",     required_argument, 0, 'Q'},
#endif
    {"no-lookup",  no_argument,       0, 'x'},
#if IDLE_OPT
    {"idle",       no_argument,       0, 'z'},
    {"no-idle",    no_argument,       0, 'Z'},
#endif
    {"duplicates", no_argument,       0, 'd'},
    {"verbose",    no_argument,       0, 'v'},
    {"help",       no_argument,       0, 'h'},
    {0, 0, 0, 0}
  };


#if defined(SIGUSR1) || defined(SIGUSR2)
static void handle_signal(int signum)
{
  switch (signum)
  {
#ifdef SIGUSR1
    case SIGUSR1:
      notify_event(save_due);
      break;
#endif
#ifdef SIGUSR2
    case SIGUSR2:
      /* Do nothing for now. */
      break;
#endif
  }
}
#endif

static time_t start_date;
#if LOG_STATS
static uint32_t init_clock, start_clock;
#endif

static int opt_ind, opt_c;

static void attribute ((noreturn)) argerror(const char *str)
{
  if (long_opts[opt_ind].val == opt_c)
    error("--%s %s: argument %s.", long_opts[opt_ind].name, optarg, str);
  else
    error("-%c %s: argument %s.", opt_c, optarg, str);
}

static uint64_t parse_uint(uint64_t limit)
{
  uint64_t num;
  uint32_t expt;
  char *tail;

  errno = 0;
  num = strtoull(optarg,&tail,0);

  if (errno == 0 && num <= limit)
  {
    switch (*tail)
    {
      case 'e':
      case 'E':
        expt = strtoul(tail+1,&tail,0);
        if (errno != 0)
          goto range_error;
        if (*tail != '\0')
          break;
        while (expt-- > 0)
          if (num > limit/10)
            goto range_error;
          else
            num *= 10;
      case '\0':
        return num;
    }

    argerror("is malformed");
  }

 range_error:
  argerror("is out of range");
}

int main(int argc, char **argv)
{
  const char *input_file_name = NULL;
#if IDLE_OPT
  int idle_opt = IDLE_DEFAULT;
#endif
#if MULTI_PATH
  int code_path = 0;
#endif
  uint32_t i, L1_opt = 0, L2_opt = 0;
  int want_help = 0;

#if LOG_STATS
  init_clock = millisec_clock();
#endif

#if HAVE_MALLOPT
  /* All memory allocation takes place during initialization, so it doesn't
     have to be fast. Reducing this threshold allows more memory to be
     returned to the system when free() is called.
  */
  mallopt(M_MMAP_THRESHOLD,16000);
#endif

#if USE_COMMAND_LINE_FILE
  /* If no comand line arguments are given, attempt to read them from file.
   */
  if (argc == 1)
    read_argc_argv(&argc,&argv,NAME COMMAND_LINE_FILE_NAME_SUFFIX);
#endif

  while ((opt_c = getopt_long(argc,argv,short_opts,long_opts,&opt_ind)) != -1)
    switch (opt_c)
    {
      case 'p':
        p_min = parse_uint(UINT64_MAX);
        break;
      case 'P':
        p_max = parse_uint(UINT64_MAX);
        break;
      case 'i':
        input_file_name = optarg;
        break;
      case 'o':
        output_file_name = optarg;
        break;
      case 'f':
        factors_file_name = optarg;
        break;
#if LEGENDRE_CACHE
      case 'C':
        cache_file_name = optarg;
        break;
#endif
      case 's':
        save_period = strtoul(optarg,NULL,0) * 60;
        break;
#if MULTI_PATH
      case '1':
        code_path = 1;
        break;
      case '2':
        code_path = 2;
        break;
#endif
      case 'l':
        L1_opt = strtoul(optarg,NULL,0);
        break;
      case 'L':
        L2_opt = strtoul(optarg,NULL,0);
        break;
#if HASHTABLE_SIZE_OPT
      case 'H':
        i = strtoul(optarg,NULL,0) * 1024;
        for (hashtable_size = 1024; hashtable_size < i; )
          hashtable_size <<= 1;
        break;
#endif
#if SUBSEQ_Q_OPT
      case 'Q':
        subseq_Q = strtoul(optarg,NULL,0);
        break;
#endif
      case 'x':
        no_lookup_opt = 1;
        break;
#if IDLE_OPT
      case 'z':
        idle_opt = 1;
        break;
      case 'Z':
        idle_opt = 0;
        break;
#endif
      case 'd':
        duplicates_opt = 1;
        break;
      case 'v':
        verbose_opt = 1;
        break;
      case 'h':
        want_help = 1;
        break;
      default:
        return 1;
    }

  banner();

  if (want_help || argc < 2)
    help();


  /* We install these signal handlers early because the default handler
     terminates the program. SIGTERM and SIGINT handlers are not installed
     until sieving begins, but termination is the right default for them.
  */
#ifdef SIGUSR1
  signal(SIGUSR1,handle_signal);
#endif
#ifdef SIGUSR2
  signal(SIGUSR2,handle_signal);
#endif

#if SUBSEQ_Q_OPT
  if (subseq_Q % BASE_MULTIPLE != 0)
    error("Subsequence base exponent Q=%"PRIu32" is not a multiple of %u.",
          subseq_Q, (unsigned int)BASE_MULTIPLE);
  if (subseq_Q != 0 && LIMIT_BASE % subseq_Q != 0)
    error("Subsequence base exponent Q=%"PRIu32" is not a divisor of %u.",
          subseq_Q, (unsigned int)LIMIT_BASE);
#endif

#if MULTI_PATH
  if (code_path == 0)
    code_path = (have_sse2()) ? 2 : 1;
  if (verbose_opt && code_path == 2)
    printf("Using SSE2 code path, ");
#endif

  set_cache_sizes(L1_opt,L2_opt);

#if LEGENDRE_CACHE
  if (no_lookup_opt != 0 && cache_file_name != NULL)
    error("--no-lookup and --cache switches are mutually exclusive.");
#endif

#if IDLE_OPT
  if (idle_opt)
    set_idle_priority();
#endif

  if (optind < argc)
    error("Unhandled argument %s.", argv[optind]);
  if (p_max == 0)
    error("--pmax is a required argument.");
  if (input_file_name == NULL)
    error("--input is a required argument.");
  if (output_file_name == NULL && factors_file_name == NULL)
    error("At least one of --output or --factors is a required argument.");

  read_newpgen_file(input_file_name);

  uint64_t lowlim = MAX(k_term,b_term);
  if (p_min <= lowlim || p_min>=p_max || p_max >= UINT64_C(1)<<MOD64_MAX_BITS)
    error("Sieve range P0 <= p <= P1 must be in %"PRIu64" < P0 < P1 < 2^%u.\n",
          lowlim, (unsigned int)MOD64_MAX_BITS);

  finish_candidate_seqs();

#if MULTI_PATH
  switch (code_path)
  {
    default:
    case 1:
      sieve();
      break;
    case 2:
      sieve_sse2();
      break;
  }
#else
  sieve();
#endif

  return EXIT_SUCCESS;
}

#define STATUS_STRFTIME_FORMAT "ETA %d %b %H:%M"
void print_status(uint64_t p, uint32_t p_per_sec)
{
  static int toggle = 0;
  char buf[32];
  double done;

  toggle = !toggle; /* toggle between reporting ETA and factor rate. */
  done = (double)(p-p_min)/(p_max-p_min);

  if (toggle && factor_count && p_per_sec)
  {
    uint64_t p_per_factor = (p-p_min)/factor_count;
    snprintf(buf,sizeof(buf),"%"PRIu64" sec/factor",p_per_factor/p_per_sec);
  }
  else if (done > 0.0)
  {
    time_t finish_date = start_date+(time(NULL)-start_date)/done;
    strftime(buf,sizeof(buf),STATUS_STRFTIME_FORMAT,localtime(&finish_date));
  }
  else
    buf[0] = '\0';

  report(0,"p=%"PRIu64", %"PRIu32" p/sec, %"PRIu32" factor%s, %.2f%% done, %s",
         p,p_per_sec,factor_count,plural(factor_count),100.0*done,buf);
}

#if !defined(NDEBUG) && USE_ASM && __i386__ && __GNUC__
#include <fenv.h>
static int old_mode;
#endif

static double expected_factors(uint32_t n, uint64_t p0, uint64_t p1)
{
  /* TODO: Use a more accurate formula. This one is only reasonable when
     p0/p1 is close to 1.
   */

  return n*(1-log(p0)/log(p1));
}

void start_srsieve(void)
{
  if (verbose_opt)
    report(1,"Expecting to find factors for about %.2f terms.",
           expected_factors(ncount,p_min,p_max));

  logger(1,"%s started: %"PRIu32" <= n <= %"PRIu32", %"PRIu64" <= p <= %"
         PRIu64, NAME, n_min, n_max, p_min, p_max);

#if !defined(NDEBUG) && USE_ASM && __i386__ && __GNUC__
  old_mode = fegetround();
#endif

  start_date = time(NULL);
#if LOG_STATS
  start_clock = millisec_clock();
#endif
}

void finish_srsieve(const char *reason, uint64_t p)
{
#if !defined(NDEBUG) && USE_ASM && __i386__ && __GNUC__
  int new_mode = fegetround();
  if (old_mode != new_mode)
    warning("FPU rounding mode was not restored: old=%d, new=%d.",
            old_mode, new_mode);
#endif

  logger(1,"%s stopped: at p=%"PRIu64" because %s.", NAME, p, reason);
  write_newpgen_file(1,p,output_file_name);

#if LOG_STATS
  uint32_t finish_clock = millisec_clock();
  logger(verbose_opt,"Sieving stats for %s, %"PRIu32" <= n <= %"PRIu32":",
         kbc_str(),n_min,n_max);
  if (finish_clock > start_clock)
    logger(verbose_opt," Range sieved: %"PRIu64" <= p <= %"PRIu64" at %"PRIu64
           " p/sec.",p_min,p,(p-p_min)*1000/(finish_clock-start_clock));
  else /* Avoid division by zero */
    logger(verbose_opt," Range sieved: %"PRIu64" <= p <= %"PRIu64".",p_min,p);
  if (factor_count > 0)
    logger(verbose_opt," Factors found: %"PRIu32" at %"PRIu32" sec/factor"
           " (expected about %.2f).",
           factor_count,(finish_clock-start_clock)/1000/factor_count,
           expected_factors(ncount,p_min,p));
  else /* Avoid division by zero */
    logger(verbose_opt," Factors found: %"PRIu32" (expected about %.2f).",
           factor_count,expected_factors(ncount,p_min,p));
  logger(verbose_opt," CPU resources used: %.2f sec (%.2f init + %.2f sieve).",
         (finish_clock-init_clock)/1000.0,
         (start_clock-init_clock)/1000.0,
         (finish_clock-start_clock)/1000.0);
#else
  logger(verbose_opt,"Found factors for %"PRIu32" term%s (expected about %.2f).",
         factor_count,plural(factor_count),expected_factors(ncount,p_min,p));
#endif
}
